import pytest

from banks import Prompt

from .conftest import anthropic_api_key_set, openai_api_key_set


@pytest.mark.e2e
@openai_api_key_set
def test_completion_openai():
    prompt_template = """
    {% set examples %}
    {% completion model="gpt-3.5-turbo-0125" %}
      {% chat role="system" %}You are a helpful assistant{% endchat %}
      {% chat role="user" %}Generate a bullet list of 3 tweets with a positive sentiment.{% endchat %}
    {% endcompletion %}
    {% endset %}

    {# output the response content #}
    Generate a tweet about the topic {{ topic }} with a positive sentiment.
    Examples:
    {{ examples }}
    """

    p = Prompt(prompt_template)
    assert "Generate a tweet about the topic LLM" in p.text({"topic": "LLM"})


@pytest.mark.e2e
@anthropic_api_key_set
def test_completion_anthropic():
    prompt_template = """
    {% set examples %}
    {% completion model="claude-3-haiku-20240307" %}
      {% chat role="system" %}You are a helpful assistant{% endchat %}
      {% chat role="user" %}Generate a bullet list of 3 tweets with a positive sentiment.{% endchat %}
    {% endcompletion %}
    {% endset %}

    {# output the response content #}
    Generate a tweet about the topic {{ topic }} with a positive sentiment.
    Examples:
    {{ examples }}
    """

    p = Prompt(prompt_template)
    assert "Generate a tweet about the topic LLM" in p.text({"topic": "LLM"})



from typing import cast

import pytest
from jinja2 import TemplateSyntaxError

from banks import Prompt
from banks.types import ContentBlock


def test_wrong_tag():
    with pytest.raises(TemplateSyntaxError):
        Prompt("{% chat %}{% endchat %}")


def test_wrong_tag_params():
    with pytest.raises(TemplateSyntaxError):
        Prompt('{% chat foo="bar" %}{% endchat %}')


def test_wrong_role_type():
    with pytest.raises(TemplateSyntaxError):
        Prompt('{% chat role="does not exist" %}{% endchat %}')


def test_blocks():
    p = Prompt('{% chat role="user" %}Tell me what you see: {{ picture | image }}{% endchat %}')
    messages = p.chat_messages({"picture": "http://foo.bar"})
    assert len(messages) == 1
    content = messages[0].content
    assert len(content) == 2
    assert content[0].type == "text"  # type: ignore
    assert content[1].type == "image_url"  # type: ignore


def test_blocks_newlines():
    p = Prompt("""{% chat role="system" %}
Some instructions:

1. Do this.
2. Then this.
3. Finally, that.
{% endchat %}""")
    messages = p.chat_messages()
    assert len(messages) == 1
    assert len(messages[0].content) == 1


def test_blocks_complex(data_path):
    p = Prompt("""
{% chat role="system" %}
Given a list of images and text from each image, please answer the question to the best of your ability.
{% endchat %}

{% chat role="user" %}
{% for image_path, text in images_and_texts %}
Here is some text: {{ text }}
And here is an image:
{{ image_path | image }}
{% endfor %}
{% endchat %}
    """)

    messages = p.chat_messages(
        {
            "images_and_texts": [
                (str(data_path / "1x1.png"), "This is the first page of the document"),
                (str(data_path / "1x1.png"), "This is the second page of the document"),
            ]
        }
    )
    assert len(messages) == 2

    blocks = cast(list[ContentBlock], messages[0].content)
    assert len(blocks) == 1
    assert blocks[0].type == "text"

    blocks = cast(list[ContentBlock], messages[1].content)
    assert len(blocks) == 4
    assert blocks[0].type == "text"
    assert blocks[1].type == "image_url"
    assert blocks[2].type == "text"
    assert blocks[3].type == "image_url"



import json
from base64 import b64encode
from pathlib import Path

import pytest

from banks import Prompt
from banks.filters.image import _is_url, image


def test_is_url():
    """Test the internal URL validation function"""
    assert _is_url("https://example.com/image.jpg") is True
    assert _is_url("http://example.com/image.jpg") is True
    assert _is_url("ftp://example.com/image.jpg") is True
    assert _is_url("not_a_url.jpg") is False
    assert _is_url("/path/to/image.jpg") is False
    assert _is_url("relative/path/image.jpg") is False
    assert _is_url("") is False
    assert _is_url("https:\\example.com/image.jpg") is False


def test_image_with_url():
    """Test image filter with a URL input"""
    url = "https://example.com/image.jpg"
    result = image(url)

    # Verify the content block wrapper
    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    # Parse the JSON content
    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "image_url"
    assert content_block["image_url"]["url"] == url


def test_image_with_file_path(tmp_path):
    """Test image filter with a file path input"""
    # Create a temporary test image file
    test_image = tmp_path / "test_image.jpg"
    test_content = b"fake image content"
    test_image.write_bytes(test_content)

    result = image(str(test_image))

    # Verify the content block wrapper
    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    # Parse the JSON content
    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "image_url"
    assert content_block["image_url"]["url"].startswith("data:image/jpeg;base64,")


def test_image_base64(tmp_path):
    """Test image filter with a binary input"""
    test_image = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAABMgA"
    result = image(test_image)

    # Verify the content block wrapper
    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    # Parse the JSON content
    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "image_url"
    assert content_block["image_url"]["url"].startswith("data:image/png;base64,")


def test_image_with_nonexistent_file():
    """Test image filter with a nonexistent file path"""
    with pytest.raises(FileNotFoundError):
        image("nonexistent/image.jpg")


def test_image_from_bytes():
    """Test image filter with bytes input"""
    png_bytes = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00"
    result = image(png_bytes)

    # Verify the content block wrapper
    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    # Parse the JSON content
    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "image_url"
    assert content_block["image_url"]["url"] == f"data:image/png;base64,{b64encode(png_bytes).decode('utf-8')}"


def test_image_from_b64_bytes():
    webp_bytes = b"RIFF\x1a\x00\x00\x00WEBPVP8 "
    b64_bytes = b64encode(webp_bytes)
    result = image(b64_bytes)

    # Verify the content block wrapper
    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    # Parse the JSON content
    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "image_url"
    assert content_block["image_url"]["url"] == f"data:image/webp;base64,{b64_bytes.decode('utf-8')}"


def test_image_content_block_structure():
    """Test the structure of the generated content block"""
    url = "https://example.com/image.jpg"
    result = image(url)

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    # Verify the content block has all expected fields
    assert set(content_block.keys()) >= {"type", "image_url"}
    assert content_block["type"] == "image_url"
    assert isinstance(content_block["image_url"], dict)
    assert "url" in content_block["image_url"]


def test_image_no_chat_block():
    here = Path(__file__).parent
    prompt = Prompt("{{ test }} and {{ another | image }}")
    messages = prompt.chat_messages({"test": "hello world", "another": str(here / "data" / "1x1.png")})
    assert len(messages) == 1
    message = messages[0]
    assert len(message.content) == 2
    assert message.content[0].text == "hello world and"
    assert message.content[1].type == "image_url"



from pathlib import Path

import pytest
import redis


def is_redis_available():
    try:
        redis.Redis(host="localhost", port=6379).ping()
        return True
    except redis.ConnectionError:
        return False


def pytest_configure(config):
    config.addinivalue_line("markers", "redis: mark test as requiring Redis")


@pytest.fixture(autouse=True)
def skip_by_redis(request):
    if request.node.get_closest_marker("redis"):
        if not is_redis_available():
            pytest.skip("Redis is not available")


@pytest.fixture
def data_path() -> Path:
    here = Path(__file__).parent
    return here / "data"



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
"""
File-based prompt registry implementation that stores all prompts in a single JSON file.

This module provides functionality to store and retrieve prompts using a single JSON file
as the storage backend. The file contains an index of all prompts with their associated
metadata and content.
"""

from __future__ import annotations

from pathlib import Path

from pydantic import BaseModel

from banks.errors import InvalidPromptError, PromptNotFoundError
from banks.prompt import Prompt, PromptModel


class PromptRegistryIndex(BaseModel):
    """
    Model representing the registry index containing all prompts.

    Stores a list of PromptModel objects that represent all prompts in the registry.
    """

    prompts: list[PromptModel] = []


class FilePromptRegistry:
    """A prompt registry storing all prompt data in a single JSON file."""

    def __init__(self, registry_index: str) -> None:
        """
        Initialize the file prompt registry.

        Args:
            registry_index: Path to the JSON file that will store the prompts

        Note:
            Creates parent directories if they don't exist.
        """
        self._index_fpath: Path = Path(registry_index)
        self._index: PromptRegistryIndex = PromptRegistryIndex(prompts=[])
        try:
            self._index = PromptRegistryIndex.model_validate_json(self._index_fpath.read_text(encoding="utf-8"))
        except FileNotFoundError:
            # init the user data folder
            self._index_fpath.parent.mkdir(parents=True, exist_ok=True)

    def get(self, *, name: str, version: str | None = None) -> Prompt:
        """
        Retrieve a prompt by name and version.

        Args:
            name: Name of the prompt to retrieve
            version: Version of the prompt (optional)

        Returns:
            The requested Prompt object

        Raises:
            PromptNotFoundError: If the requested prompt doesn't exist
        """
        _, model = self._get_prompt_model(name, version)
        return Prompt(**model.model_dump())

    def set(self, *, prompt: Prompt, overwrite: bool = False) -> None:
        """
        Store a prompt in the registry.

        Args:
            prompt: The Prompt object to store
            overwrite: Whether to overwrite an existing prompt

        Raises:
            InvalidPromptError: If prompt exists and overwrite=False
        """
        try:
            idx, p_model = self._get_prompt_model(prompt.name, prompt.version)
            if overwrite:
                self._index.prompts[idx] = PromptModel.from_prompt(prompt)
                self._save()
            else:
                msg = f"Prompt with name '{prompt.name}' already exists. Use overwrite=True to overwrite"
                raise InvalidPromptError(msg)
        except PromptNotFoundError:
            p_model = PromptModel.from_prompt(prompt)
            self._index.prompts.append(p_model)
            self._save()

    def _save(self) -> None:
        """
        Save the prompt index to the JSON file.

        Writes the current state of the registry to disk.
        """
        with open(self._index_fpath, "w", encoding="utf-8") as f:
            f.write(self._index.model_dump_json())

    def _get_prompt_model(self, name: str | None, version: str | None) -> tuple[int, PromptModel]:
        """
        Find a prompt model in the index by name and version.

        Args:
            name: Name of the prompt
            version: Version of the prompt

        Returns:
            Tuple of (index position, PromptModel)

        Raises:
            PromptNotFoundError: If the prompt doesn't exist in the index
        """
        for i, model in enumerate(self._index.prompts):
            if model.name == name and model.version == version:
                return i, model

        msg = f"cannot find prompt with name '{name}' and version '{version}'"
        raise PromptNotFoundError(msg)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT

from jinja2 import TemplateSyntaxError, nodes
from jinja2.ext import Extension

from banks.types import chat_message_from_text

SUPPORTED_TYPES = ("system", "user", "assistant")


class ChatExtension(Extension):
    """
    `chat` can be used to render prompt text as structured ChatMessage objects.

    Example:
        ```
        {% chat role="system" %}
        You are a helpful assistant.
        {% endchat %}
        ```
    """

    # a set of names that trigger the extension.
    tags = {"chat"}  # noqa

    def parse(self, parser):
        # We get the line number of the first token for error reporting
        lineno = next(parser.stream).lineno

        # Gather tokens up to the next block_end ('%}')
        gathered = []
        while parser.stream.current.type != "block_end":
            gathered.append(next(parser.stream))

        # If all has gone well, we will have one triplet of tokens:
        #   (type='name, value='role'),
        #   (type='assign', value='='),
        #   (type='string', value='user'),
        # Anything else is a parse error
        error_msg = f"Invalid syntax for chat attribute, got '{gathered}', expected role=\"value\""
        try:
            attr_name, attr_assign, attr_value = gathered  # pylint: disable=unbalanced-tuple-unpacking
        except ValueError:
            raise TemplateSyntaxError(error_msg, lineno) from None

        # Validate tag attributes
        if attr_name.value != "role" or attr_assign.value != "=":
            raise TemplateSyntaxError(error_msg, lineno)

        if attr_value.value not in SUPPORTED_TYPES:
            types = ", ".join(SUPPORTED_TYPES)
            msg = f"Unknown role type '{attr_value.value}', use one of ({types})"
            raise TemplateSyntaxError(msg, lineno)

        # Pass the role name to the CallBlock node
        args: list[nodes.Expr] = [nodes.Const(attr_value.value)]

        # Message body
        body = parser.parse_statements(("name:endchat",), drop_needle=True)

        # Build messages list
        return nodes.CallBlock(self.call_method("_store_chat_messages", args), [], [], body).set_lineno(lineno)

    def _store_chat_messages(self, role, caller):
        """
        Helper callback.
        """
        cm = chat_message_from_text(role=role, content=caller())
        return cm.model_dump_json(exclude_none=True) + "\n"



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import re
from pathlib import Path
from typing import cast
from urllib.parse import urlparse

import filetype  # type: ignore[import-untyped]
from filetype.types.video import IsoBmff  # type: ignore[import-untyped]

from banks.types import ContentBlock, InputVideo, VideoFormat, resolve_binary

BASE64_VIDEO_REGEX = re.compile(r"video\/.*;base64,.*")


class M3gp(IsoBmff):
    """
    Implements the 3gp video type matcher.

    The type matcher in the filetype lib does not work correctly for 3gp files,
    so implement our own here.
    """

    MIME = "video/3gpp"
    EXTENSION = "3gp"

    def __init__(self):
        super().__init__(mime=M3gp.MIME, extension=M3gp.EXTENSION)

    def match(self, buf):
        if not self._is_isobmff(buf):
            return False

        major_brand, _, compatible_brands = self._get_ftyp(buf)
        for brand in compatible_brands:
            if brand in ["3gp4", "3gp5", "3gpp"]:
                return True
        return major_brand in ["3gp4", "3gp5", "3gpp"]


def _is_url(string: str) -> bool:
    """Check if a string is a URL."""
    result = urlparse(string)
    if not result.scheme:
        return False

    if not result.netloc:
        # The only valid format when netloc is empty is base64 data urls
        return all([result.scheme == "data", BASE64_VIDEO_REGEX.match(result.path)])

    return True


def _get_video_format_from_url(url: str) -> VideoFormat:
    """Extract video format from URL.

    Tries to determine format from URL path or defaults to mp4.
    """
    parsed = urlparse(url)
    path = parsed.path.lower()

    # Based on formats supported by Gemini https://ai.google.dev/gemini-api/docs/video-understanding
    for fmt in ("mp4", "mpeg", "mov", "avi", "flv", "mpg", "webm", "wmv", "3gpp"):
        if path.endswith(f".{fmt}"):
            return cast(VideoFormat, fmt)
    # Default to mp4 if format cannot be determined
    return "mp4"


def _get_video_format_from_bytes(data: bytes) -> VideoFormat:
    """Extract video format from bytes data using filetype library."""
    m3gp = M3gp()
    if m3gp not in filetype.types:
        filetype.add_type(m3gp)

    kind = filetype.guess(data)
    if kind is not None:
        fmt = kind.extension
        if fmt in ("mp4", "mpg", "mov", "avi", "flv", "webm", "wmv", "3gp"):
            return cast(VideoFormat, fmt)
    # Default to mp4 if format cannot be determined
    return "mp4"


def video(value: str | bytes) -> str:
    """Wrap the filtered value into a ContentBlock of type video.

    The resulting ChatMessage will have the field `content` populated with a list of ContentBlock objects.

    Supports both file paths and URLs (including data URLs).

    Example:
        ```jinja
        {{ "path/to/video/file.mp4" | video }}
        {{ "https://example.com/video.mp4" | video }}
        ```
    """
    if isinstance(value, bytes):
        video_format = _get_video_format_from_bytes(resolve_binary(value, as_base64=False))
        input_video = InputVideo.from_bytes(value, video_format=video_format)
    elif _is_url(value):
        video_format = _get_video_format_from_url(value)
        input_video = InputVideo.from_url(value, video_format)
    else:
        input_video = InputVideo.from_path(Path(value))
    block = ContentBlock.model_validate({"type": "video", "input_video": input_video})
    return f"<content_block>{block.model_dump_json()}</content_block>"



from pathlib import Path

from platformdirs import user_data_path

from banks.config import _BanksConfig


def test_config_defaults():
    c = _BanksConfig()
    assert c.ASYNC_ENABLED is False
    assert c.USER_DATA_PATH == user_data_path("banks")


def test_config_env_override(monkeypatch):
    c = _BanksConfig()
    monkeypatch.setenv("BANKS_ASYNC_ENABLED", "true")
    assert c.ASYNC_ENABLED is True
    monkeypatch.setenv("BANKS_ASYNC_ENABLED", "false")
    assert c.ASYNC_ENABLED is False
    monkeypatch.setenv("BANKS_USER_DATA_PATH", "/")
    assert c.USER_DATA_PATH == Path("/")

    class TestConfig(_BanksConfig):
        FOO: int = 0

    c = TestConfig()
    assert c.FOO == 0
    monkeypatch.setenv("BANKS_FOO", "42")
    assert c.FOO == 42


def test_config_env_prefix(monkeypatch):
    c = _BanksConfig("BANKS_TEST_")
    monkeypatch.setenv("BANKS_ASYNC_ENABLED", "true")
    assert c.ASYNC_ENABLED is False
    monkeypatch.setenv("BANKS_TEST_ASYNC_ENABLED", "true")
    assert c.ASYNC_ENABLED is True



import json
from base64 import b64encode
from pathlib import Path

import pytest

from banks import Prompt
from banks.filters.video import _get_video_format_from_bytes, _get_video_format_from_url, _is_url, video


@pytest.fixture
def empty_mov():
    here = Path(__file__).parent
    return here / "data" / "empty.mov"


def test_video_with_file_path(empty_mov):
    """Test video filter with a file path input"""
    result = video(str(empty_mov))

    # Verify the content block wrapper
    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    # Parse the JSON content
    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "video"
    assert content_block["input_video"]["format"].startswith("mov")


def test_video_with_nonexistent_file():
    """Test video filter with a nonexistent file path"""
    with pytest.raises(FileNotFoundError):
        video("nonexistent/video.mov")


def test_video_with_url():
    """Test video filter with a URL input (no filesystem access)."""
    url = "https://example.com/video.webm"
    result = video(url)

    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "video"
    assert content_block["input_video"]["data"] == url
    assert content_block["input_video"]["format"] == "webm"


def test_is_url_variants():
    assert _is_url("relative/path.mov") is False
    assert _is_url("https://example.com/video.mov") is True
    assert _is_url("data:video/mov;base64,AAAA") is True
    assert _is_url("data:text/plain;base64,AAAA") is False


def test_get_video_format_from_url():
    assert _get_video_format_from_url("https://example.com/video.MOV") == "mov"
    assert _get_video_format_from_url("https://example.com/video") == "mp4"


def test_get_video_from_bytes():
    mp4_bytes = b"\x00\x00\x00\x18ftypmp42"
    result = video(mp4_bytes)

    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "video"
    assert content_block["input_video"]["format"] == "mp4"
    assert content_block["input_video"]["data"] == b64encode(mp4_bytes).decode("utf-8")


def test_get_video_from_b64_bytes():
    webm_bytes = b"\x1a\x45\xdf\xa3\x42\x82\x84webm"
    result = video(webm_bytes)

    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "video"
    assert content_block["input_video"]["format"] == "webm"
    assert content_block["input_video"]["data"] == b64encode(webm_bytes).decode("utf-8")


def test_get_video_format_from_bytes():
    mp4 = b"\x00\x00\x00\x18ftypmp42" + b"\x00" * (0x18 - len(b"\x00\x00\x00\x18ftypmp42"))
    mpeg = b"\x00\x00\x01\xba"
    mov = b"\x00\x00\x00\x14ftypqt  " + b"\x00" * (0x14 - len(b"\x00\x00\x00\x14ftypqt  "))
    avi = b"RIFF$\x00\x00\x00AVI LIST"
    flv = b"FLV\x01\x05\x00\x00\x00\x09"
    webm = b"\x1a\x45\xdf\xa3\x42\x82\x84webm"
    wmv = bytes.fromhex("3026B2758E66CF11A6D900AA0062CE6C")
    gpp = b"\x00\x00\x00\x18ftyp3gp5\x00\x00\x00\x00isom3gp4"
    unknown = b"\x00\x01\x02\x03\x04\x05"

    assert _get_video_format_from_bytes(mp4) == "mp4"
    assert _get_video_format_from_bytes(mpeg) == "mpg"
    assert _get_video_format_from_bytes(mov) == "mov"
    assert _get_video_format_from_bytes(avi) == "avi"
    assert _get_video_format_from_bytes(flv) == "flv"
    assert _get_video_format_from_bytes(webm) == "webm"
    assert _get_video_format_from_bytes(wmv) == "wmv"
    assert _get_video_format_from_bytes(gpp) == "3gp"
    assert _get_video_format_from_bytes(unknown) == "mp4"  # Default


def test_video_no_chat_block(empty_mov):
    prompt = Prompt("{{ test }} and {{ another | video }}")
    messages = prompt.chat_messages({"test": "hello world", "another": str(empty_mov)})
    assert len(messages) == 1
    message = messages[0]
    assert len(message.content) == 2
    assert message.content[0].text == "hello world and"  # type: ignore
    assert message.content[1].type == "video"  # type:ignore



import json
import xml.etree.ElementTree as ET
from typing import Any, Optional, Union
from xml.dom.minidom import parseString

from pydantic import BaseModel


def _deserialize(string: str) -> Optional[dict]:
    try:
        return json.loads(string)
    except json.JSONDecodeError:
        return None


def _prepare_dictionary(value: Union[str, BaseModel, dict[str, Any]]):
    root_tag = "input"
    if isinstance(value, str):
        model: Optional[dict[str, Any]] = _deserialize(value)
        if model is None:
            msg = f"{value} is not deserializable"
            raise ValueError(msg)
    elif isinstance(value, BaseModel):
        model = value.model_dump()
        root_tag = value.__class__.__name__.lower()
    elif isinstance(value, dict):
        model = value.copy()
        for k in value.keys():
            if not isinstance(k, str):
                key = str(k)
                if isinstance(k, (int, float)):
                    key = "_" + key
                v = model.pop(k)
                model[key.lower()] = v
    else:
        msg = f"Input can only be of type BaseModel, dictionary or deserializable string. Got {type(value)}"
        raise ValueError(msg)
    return model, root_tag


def xml(value: Union[str, BaseModel, dict[str, Any]]) -> str:
    """
    Convert a Pydantic model, a deserializable string or a dictionary into an XML string.

    Example:
        ```jinja
        {{'{"username": "user", "email": "example@email.com"}' | to_xml}}
        "
        <input>
            <username>user</username>
            <email>example@email.com</email>
        </input>
        "
        ```
    """
    model, root_tag = _prepare_dictionary(value)
    xml_model = ET.Element(root_tag)
    for k, v in model.items():
        sub = ET.SubElement(xml_model, k)
        sub.text = str(v)
    xml_str = ET.tostring(xml_model, encoding="unicode")
    return parseString(xml_str).toprettyxml().replace('<?xml version="1.0" ?>\n', "")  # noqa: S318



import secrets

from griffe import Docstring, parse_google, parse_numpy, parse_sphinx


def strtobool(val: str) -> bool:
    """
    Convert a string representation of truth to True or False.

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True

    if val in ("n", "no", "f", "false", "off", "0"):
        return False

    msg = f"invalid truth value {val}"
    raise ValueError(msg)


def generate_canary_word(prefix: str = "BANKS[", suffix: str = "]", token_length: int = 8) -> str:
    return f"{prefix}{secrets.token_hex(token_length // 2)}{suffix}"


def python_type_to_jsonschema(python_type: type) -> str:
    """Given a Python type, returns the jsonschema string describing it."""
    if python_type is str:
        return "string"
    if python_type is int:
        return "integer"
    if python_type is float:
        return "number"
    if python_type is bool:
        return "boolean"
    if python_type is list:
        return "array"
    if python_type is dict:
        return "object"

    msg = f"Unsupported type: {python_type}"
    raise ValueError(msg)


def parse_params_from_docstring(docstring: str) -> dict[str, dict[str, str]]:
    param_docs = []
    ds = Docstring(docstring)
    for parser in (parse_google, parse_numpy, parse_sphinx):
        sections = parser(ds)
        for section in sections:
            if section.kind == "parameters":
                param_docs = section.value
                break
        if param_docs:
            break

    ret = {}
    for d in param_docs:
        d_dict = d.as_dict()
        ret[d_dict.pop("name")] = d_dict

    return ret



import pytest
import regex as re

from banks.utils import generate_canary_word, parse_params_from_docstring, python_type_to_jsonschema, strtobool


def test_generate_canary_word_defaults():
    default = generate_canary_word()
    assert re.match(r"BANKS\[.{8}\]", default)


def test_generate_canary_word_params():
    only_token = generate_canary_word(prefix="", suffix="", token_length=16)
    assert re.match(r".{16}", only_token)

    only_prefix = generate_canary_word(prefix="foo", suffix="")
    assert re.match(r"foo.{8}", only_prefix)

    only_suffix = generate_canary_word(prefix="", suffix="foo")
    assert re.match(r".{8}foo", only_suffix)


def test_strtobool_error():
    with pytest.raises(ValueError):
        strtobool("42")


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ("y", True),
        ("yes", True),
        ("t", True),
        ("true", True),
        ("on", True),
        ("1", True),
        ("n", False),
        ("no", False),
        ("f", False),
        ("false", False),
        ("off", False),
        ("0", False),
        pytest.param("42", True, marks=pytest.mark.xfail),
    ],
)
def test_strtobool(test_input, expected):
    assert strtobool(test_input) == expected


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (type("I am a string"), "string"),  # noqa
        (type(42), "integer"),  # noqa
        (type(0.42), "number"),  # noqa
        (type(True), "boolean"),  # noqa
        (type([]), "array"),
        (type({"foo": "bar"}), "object"),
    ],
)
def test_python_type_to_jsonschema(test_input, expected):
    assert python_type_to_jsonschema(test_input) == expected
    with pytest.raises(ValueError, match="Unsupported type: <class 'type'>"):
        python_type_to_jsonschema(type(Exception))


def test_parse_params_from_docstring_google():
    def my_test_function(test_param: str):
        """A docstring.

        Args:
            test_param (str): The test parameter.
        """
        pass

    assert parse_params_from_docstring(my_test_function.__doc__) == {  # type: ignore
        "test_param": {"annotation": "str", "description": "The test parameter."}
    }


def test_parse_params_from_docstring_numpy():
    def my_test_function(test_param: str):
        """A docstring.

        Parameters
        ----------
        test_param : str
            The test parameter.
        """
        pass

    assert parse_params_from_docstring(my_test_function.__doc__) == {  # type: ignore
        "test_param": {"annotation": "str", "description": "The test parameter."}
    }


def test_parse_params_from_docstring_sphinx():
    def my_test_function(test_param: str):
        """A docstring.

        :param test_param: The test parameter.
        :type test_param: str
        """
        pass

    assert parse_params_from_docstring(my_test_function.__doc__) == {  # type: ignore
        "test_param": {"annotation": "str", "description": "The test parameter."}
    }


def test_parse_params_from_docstring_empty():
    assert parse_params_from_docstring("") == {}



from os import getenv
from unittest import mock

import pytest
from jinja2.environment import Environment
from litellm.types.utils import ChatCompletionMessageToolCall, Function

from banks.errors import InvalidPromptError, LLMError
from banks.extensions.completion import CompletionExtension
from banks.types import ChatMessage, Tool


@pytest.fixture
def ext():
    return CompletionExtension(environment=Environment())


@pytest.fixture
def mocked_choices_no_tools():
    return [mock.MagicMock(message=mock.MagicMock(tool_calls=None, content="some response"))]


@pytest.fixture
def mocked_choices_with_tools():
    return [
        mock.MagicMock(
            message=mock.MagicMock(
                tool_calls=[
                    ChatCompletionMessageToolCall(
                        id="call_DN6IiLULWZw7sobV6puCji1O",
                        function=Function(
                            arguments='{"location": "San Francisco", "unit": "celsius"}', name="get_current_weather"
                        ),
                        type="function",
                    ),
                    ChatCompletionMessageToolCall(
                        id="call_ERm1JfYO9AFo2oEWRmWUd40c",
                        function=Function(
                            arguments='{"location": "Tokyo", "unit": "celsius"}', name="get_current_weather"
                        ),
                        type="function",
                    ),
                    ChatCompletionMessageToolCall(
                        id="call_2lvUVB1y4wKunSxTenR0zClP",
                        function=Function(
                            arguments='{"location": "Paris", "unit": "celsius"}', name="get_current_weather"
                        ),
                        type="function",
                    ),
                ],
                content="some response",
            )
        )
    ]


@pytest.fixture
def tools():
    return [
        Tool.model_validate(
            {
                "type": "function",
                "function": {
                    "name": "getenv",
                    "description": "Get an environment variable, return None if it doesn't exist.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "key": {
                                "type": "string",
                                "description": "The name of the environment variable",
                            },
                            "default": {
                                "type": "string",
                                "description": "The value to return if the variable was not found",
                            },
                        },
                        "required": ["key"],
                    },
                },
                "import_path": "os.getenv",
            }
        )
    ]


def test__body_to_messages(ext):
    assert ext._body_to_messages(' \n{"role":"user", "content":"hello"}') == (
        [ChatMessage(role="user", content="hello")],
        [],
    )
    assert ext._body_to_messages('{"role":"user", "content":"hello"}\n HELLO!') == (
        [ChatMessage(role="user", content="hello")],
        [],
    )
    with pytest.raises(InvalidPromptError, match="Completion must contain at least one chat message"):
        ext._body_to_messages(" ")
    with pytest.raises(InvalidPromptError, match="Completion must contain at least one chat message"):
        ext._body_to_messages(" \nhello\n ")


def test__do_completion_no_prompt(ext):
    with pytest.raises(InvalidPromptError, match="Completion must contain at least one chat message"):
        ext._do_completion("test-model", lambda: " ")


@pytest.mark.asyncio
async def test__do_completion_async_no_prompt(ext):
    with pytest.raises(InvalidPromptError, match="Completion must contain at least one chat message"):
        await ext._do_completion_async("test-model", lambda: " ")


def test__do_completion_no_tools(ext, mocked_choices_no_tools):
    with mock.patch("litellm.completion") as mocked_completion:
        mocked_completion.return_value.choices = mocked_choices_no_tools
        ext._do_completion("test-model", lambda: '{"role":"user", "content":"hello"}')
        mocked_completion.assert_called_with(
            model="test-model", messages=[ChatMessage(role="user", content="hello").model_dump()], tools=None
        )


@pytest.mark.asyncio
async def test__do_completion_async_no_tools(ext, mocked_choices_no_tools):
    with mock.patch("litellm.acompletion") as mocked_completion:
        mocked_completion.return_value.choices = mocked_choices_no_tools
        await ext._do_completion_async("test-model", lambda: '{"role":"user", "content":"hello"}')
        mocked_completion.assert_called_with(
            model="test-model",
            messages=[{"role": "user", "content": "hello", "tool_call_id": None, "name": None}],
            tools=None,
        )


def test__do_completion_with_tools(ext, mocked_choices_with_tools):
    ext._get_tool_callable = mock.MagicMock(return_value=lambda location, unit: f"I got {location} with {unit}")
    ext._body_to_messages = mock.MagicMock(
        return_value=(
            [ChatMessage(role="user", content="message1"), ChatMessage(role="user", content="message2")],
            [mock.MagicMock(), mock.MagicMock()],
        )
    )
    with mock.patch("litellm.completion") as mocked_completion:
        mocked_completion.return_value.choices = mocked_choices_with_tools
        ext._do_completion("test-model", lambda: '{"role":"user", "content":"hello"}')
        calls = mocked_completion.call_args_list
        assert len(calls) == 2  # complete query, complete with tool results
        assert len(calls[0].kwargs["tools"]) == 2
        for m in calls[1].kwargs["messages"]:
            if type(m) is ChatMessage:
                assert m.role == "tool"
                assert m.name == "get_current_weather"


@pytest.mark.asyncio
async def test__do_completion_async_with_tools(ext, mocked_choices_with_tools, tools):
    ext._get_tool_callable = mock.MagicMock(return_value=lambda location, unit: f"I got {location} with {unit}")
    ext._body_to_messages = mock.MagicMock(
        return_value=(
            [ChatMessage(role="user", content="message1"), ChatMessage(role="user", content="message2")],
            tools,
        )
    )
    with mock.patch("litellm.acompletion") as mocked_completion:
        mocked_completion.return_value.choices = mocked_choices_with_tools
        await ext._do_completion_async("test-model", lambda: '{"role":"user", "content":"hello"}')
        calls = mocked_completion.call_args_list
        assert len(calls) == 2  # complete query, complete with tool results
        assert calls[0].kwargs["tools"] == [t.model_dump() for t in tools]
        for m in calls[1].kwargs["messages"]:
            if type(m) is ChatMessage:
                assert m.role == "tool"
                assert m.name == "get_current_weather"


def test__do_completion_with_tools_malformed(ext, mocked_choices_with_tools):
    mocked_choices_with_tools[0].message.tool_calls[0].function.name = None
    with mock.patch("litellm.completion") as mocked_completion:
        mocked_completion.return_value.choices = mocked_choices_with_tools
        with pytest.raises(LLMError):
            ext._do_completion("test-model", lambda: '{"role":"user", "content":"hello"}')


@pytest.mark.asyncio
async def test__do_completion_async_with_tools_malformed(ext, mocked_choices_with_tools):
    mocked_choices_with_tools[0].message.tool_calls[0].function.name = None
    with mock.patch("litellm.acompletion") as mocked_completion:
        mocked_completion.return_value.choices = mocked_choices_with_tools
        with pytest.raises(LLMError):
            await ext._do_completion_async("test-model", lambda: '{"role":"user", "content":"hello"}')


@pytest.mark.asyncio
async def test__do_completion_async_no_prompt_no_tools(ext, mocked_choices_no_tools):
    with mock.patch("litellm.acompletion") as mocked_completion:
        mocked_completion.return_value.choices = mocked_choices_no_tools
        await ext._do_completion_async("test-model", lambda: '{"role":"user", "content":"hello"}')
        mocked_completion.assert_called_with(
            model="test-model",
            messages=[{"role": "user", "content": "hello", "tool_call_id": None, "name": None}],
            tools=None,
        )


def test__get_tool_callable(ext, tools):
    tool_call = mock.MagicMock()

    tool_call.function.name = "getenv"
    assert ext._get_tool_callable(tools, tool_call) == getenv

    tool_call.function.name = "another_func"
    with pytest.raises(ValueError, match="Function another_func not found in available tools"):
        ext._get_tool_callable(tools, tool_call)



import base64
from pathlib import Path

import pytest

from banks.types import ImageUrl, InputAudio


def test_image_url_from_base64():
    """Test creating ImageUrl from base64 encoded data"""
    test_data = "Hello, World!"
    base64_data = base64.b64encode(test_data.encode()).decode("utf-8")
    media_type = "image/jpeg"

    image_url = ImageUrl.from_base64(media_type, base64_data)
    expected_url = f"data:{media_type};base64,{base64_data}"
    assert image_url.url == expected_url


def test_image_url_from_path(tmp_path):
    """Test creating ImageUrl from a file path"""
    # Create a temporary test image file
    test_image = tmp_path / "test_image.jpg"
    test_content = b"fake image content"
    test_image.write_bytes(test_content)

    image_url = ImageUrl.from_path(test_image)

    # Verify the URL starts with the expected data URI prefix
    assert image_url.url.startswith("data:image/jpeg;base64,")

    # Decode the base64 part and verify the content matches
    base64_part = image_url.url.split(",")[1]
    decoded_content = base64.b64decode(base64_part)
    assert decoded_content == test_content


def test_image_url_from_path_nonexistent():
    """Test creating ImageUrl from a nonexistent file path"""
    with pytest.raises(FileNotFoundError):
        ImageUrl.from_path(Path("nonexistent.jpg"))


def test_input_audio_from_path(tmp_path):
    """Test creating InputAudio from a file path"""
    # Create a temporary test image file
    test_audio = tmp_path / "test_audio.wav"
    test_content = b"fake audio data"
    test_audio.write_bytes(test_content)

    input_audio = InputAudio.from_path(test_audio)

    assert input_audio.format == "wav"
    decoded_content = base64.b64decode(input_audio.data)
    assert decoded_content == test_content


def test_input_audio_from_path_nonexistent():
    """Test creating ImageUrl from a nonexistent file path"""
    with pytest.raises(FileNotFoundError):
        InputAudio.from_path(Path("nonexistent.wav"))



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from .directory import DirectoryPromptRegistry
from .file import FilePromptRegistry

__all__ = ("FilePromptRegistry", "DirectoryPromptRegistry")



from pathlib import Path
from unittest import mock

import pytest
import regex as re
from jinja2 import Environment

from banks import AsyncPrompt, Prompt
from banks.cache import DefaultCache
from banks.errors import AsyncError


def test_canary_word_generation():
    p = Prompt("{{canary_word}}This is my prompt")
    assert re.match(r"BANKS\[.{8}\]This is my prompt", p.text())


def test_canary_word_leaked():
    p = Prompt("{{canary_word}}This is my prompt")
    assert p.canary_leaked(p.text())

    p = Prompt("This is my prompt")
    assert not p.canary_leaked(p.text())


def test_prompt_cache():
    mock_cache = DefaultCache()
    mock_cache.set = mock.Mock()
    p = Prompt("This is my prompt", render_cache=mock_cache)
    p.text()
    mock_cache.set.assert_called_once()


def test_ctor():
    p = Prompt(
        text="This is raw text",
        version="1.0",
        metadata={"LLM": "GPT-3.5"},
        name="test_prompt",
        canary_word="FOO",
        render_cache=DefaultCache(),
    )
    assert p.raw == "This is raw text"
    assert p.version == "1.0"
    assert p.metadata == {"LLM": "GPT-3.5"}
    assert p.name == "test_prompt"
    assert p.canary_leaked("The message is FOO")
    assert p.text() == "This is raw text"
    assert p.text() == "This is raw text"


def test_ctor_async_disabled():
    with mock.patch("banks.prompt.config", ASYNC_ENABLED=False):
        with pytest.raises(AsyncError):
            AsyncPrompt(text="This is raw text")


@pytest.mark.asyncio
async def test_ctor_async():
    with mock.patch("banks.prompt.config", ASYNC_ENABLED=True):
        p = AsyncPrompt(
            text="This is raw text",
            version="1.0",
            metadata={"LLM": "GPT-3.5"},
            canary_word="FOO",
            render_cache=DefaultCache(),
        )
        p._template = Environment(autoescape=True, enable_async=True).from_string(p.raw)
        assert p.raw == "This is raw text"
        assert p.version == "1.0"
        assert p.metadata == {"LLM": "GPT-3.5"}
        assert p.canary_leaked("The message is FOO")
        assert await p.text() == "This is raw text"
        assert await p.text() == "This is raw text"


def test__get_context():
    p = Prompt(text="This is raw text")
    assert p._get_context(None) == p.defaults
    data = {"foo": 42}
    assert p._get_context(data) == data | p.defaults


def test_chat_messages():
    p_file = Path(__file__).parent / "templates" / "chat.jinja"
    p = Prompt(p_file.read_text())

    assert (
        p.text()
        == """
{"role":"system","content":[{"type":"text","text":"You are a helpful assistant."}]}

{"role":"user","content":[{"type":"text","cache_control":{"type":"ephemeral"},"text":"Hello, <bold>how are you?</bold>"}]}

{"role":"system","content":[{"type":"text","text":"I'm doing well, thank you! How can I assist you today?"}]}

{"role":"user","content":[{"type":"text","text":"Can you explain quantum computing?"}]}

Some random text.
""".strip()
    )

    messages = p.chat_messages()
    assert len(messages) == 4
    assert messages[3].role == "user"
    assert messages[3].content[0].text == "Can you explain quantum computing?"


def test_chat_messages_cached():
    mock_cache = DefaultCache()
    mock_cache.set = mock.Mock()
    p_file = Path(__file__).parent / "templates" / "chat.jinja"
    p = Prompt(p_file.read_text(), render_cache=mock_cache)
    p.chat_messages()
    mock_cache.set.assert_called_once()


def test_chat_message_no_chat_tag():
    text = "This is raw text"
    p = Prompt(text=text)
    assert len(p.chat_messages()) == 1
    assert len(p.chat_messages()[0].content) == 1
    assert p.chat_messages()[0].content[0].type == "text"
    assert p.chat_messages()[0].content[0].text == "This is raw text"


def test_variables():
    prompt_text = "This is a {{ first_variable.content }} and this is {{ another_variable }}"
    p = Prompt(text=prompt_text)
    assert p.variables == {"another_variable", "first_variable"}



import json
from base64 import b64encode
from pathlib import Path

import pytest

from banks import Prompt
from banks.filters.audio import _get_audio_format_from_bytes, _get_audio_format_from_url, _is_url, audio


@pytest.fixture
def empty_wav():
    here = Path(__file__).parent
    return here / "data" / "empty.wav"


def test_audio_with_file_path(empty_wav):
    """Test audio filter with a file path input"""
    result = audio(str(empty_wav))

    # Verify the content block wrapper
    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    # Parse the JSON content
    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "audio"
    assert content_block["input_audio"]["format"].startswith("wav")


def test_audio_with_nonexistent_file():
    """Test audio filter with a nonexistent file path"""
    with pytest.raises(FileNotFoundError):
        audio("nonexistent/audio.wav")


def test_audio_with_url():
    """Test audio filter with a URL input (no filesystem access)."""
    url = "https://example.com/sound.ogg"
    result = audio(url)

    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "audio"
    assert content_block["input_audio"]["data"] == url
    assert content_block["input_audio"]["format"] == "ogg"


def test_is_url_variants():
    assert _is_url("relative/path.wav") is False
    assert _is_url("https://example.com/sound.wav") is True
    assert _is_url("data:audio/wav;base64,AAAA") is True
    assert _is_url("data:text/plain;base64,AAAA") is False


def test_get_audio_format_from_url():
    assert _get_audio_format_from_url("https://example.com/sound.WAV") == "wav"
    assert _get_audio_format_from_url("https://example.com/sound") == "mp3"


def test_audio_from_bytes():
    mp3 = b"\xff\xfb\x90\x64\x00\x0f\xff\xf3\xb4\x00\x00\x00\x00\x00\x00\x00"
    result = audio(mp3)

    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)
    assert content_block["type"] == "audio"
    assert content_block["input_audio"]["data"] == b64encode(mp3).decode("utf-8")
    assert content_block["input_audio"]["format"] == "mp3"


def test_audio_from_b64_bytes():
    wav = b"RIFF$\x00\x00\x00WAVEfmt "
    b64_wav = b64encode(wav)
    result = audio(b64_wav)

    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)
    assert content_block["type"] == "audio"
    assert content_block["input_audio"]["data"] == b64_wav.decode("utf-8")
    assert content_block["input_audio"]["format"] == "wav"


def test_get_audio_format_from_bytes():
    wav = b"RIFF$\x00\x00\x00WAVEfmt "  # Beginning bytes of a WAV file
    mp3 = b"\xff\xfb\x90\x64\x00\x0f\xff\xf3\xb4\x00\x00\x00\x00\x00\x00\x00"
    m4a = b"\x00\x00\x00\x20ftypM4A "  # Beginning bytes of an M4A file
    webm = b"\x1a\x45\xdf\xa3\x42\x82\x84webm"  # Beginning bytes of a WebM file
    ogg = b"OggS"  # Beginning bytes of an OGG file
    flac = b"fLaC"  # Beginning bytes of a FLAC file
    undetermined = b"\x00\x01\x02\x03\x04\x05"

    assert _get_audio_format_from_bytes(wav) == "wav"
    assert _get_audio_format_from_bytes(mp3) == "mp3"
    assert _get_audio_format_from_bytes(m4a) == "m4a"
    assert _get_audio_format_from_bytes(webm) == "webm"
    assert _get_audio_format_from_bytes(ogg) == "ogg"
    assert _get_audio_format_from_bytes(flac) == "flac"
    assert _get_audio_format_from_bytes(undetermined) == "mp3"


def test_audio_no_chat_block(empty_wav):
    prompt = Prompt("{{ test }} and {{ another | audio }}")
    messages = prompt.chat_messages({"test": "hello world", "another": str(empty_wav)})
    assert len(messages) == 1
    message = messages[0]
    assert len(message.content) == 2
    assert message.content[0].text == "hello world and"  # type: ignore
    assert message.content[1].type == "audio"  # type:ignore



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import re
from pathlib import Path
from typing import cast
from urllib.parse import urlparse

import filetype  # type: ignore[import-untyped]

from banks.types import AudioFormat, ContentBlock, InputAudio, resolve_binary

BASE64_AUDIO_REGEX = re.compile(r"audio\/.*;base64,.*")


def _is_url(string: str) -> bool:
    """Check if a string is a URL."""
    result = urlparse(string)
    if not result.scheme:
        return False

    if not result.netloc:
        # The only valid format when netloc is empty is base64 data urls
        return all([result.scheme == "data", BASE64_AUDIO_REGEX.match(result.path)])

    return True


def _get_audio_format_from_url(url: str) -> AudioFormat:
    """Extract audio format from URL.

    Tries to determine format from URL path or defaults to mp3.
    """
    parsed = urlparse(url)
    path = parsed.path.lower()
    for fmt in ("mp3", "wav", "m4a", "webm", "ogg", "flac"):
        if path.endswith(f".{fmt}"):
            return cast(AudioFormat, fmt)
    # Default to mp3 if format cannot be determined
    return "mp3"


def _get_audio_format_from_bytes(data: bytes) -> AudioFormat:
    """Extract audio format from bytes data using filetype library."""
    kind = filetype.guess(data)
    if kind is not None:
        fmt = kind.extension
        if fmt in ("mp3", "wav", "m4a", "webm", "ogg", "flac"):
            return cast(AudioFormat, fmt)
    # Default to mp3 if format cannot be determined
    return "mp3"


def audio(value: str | bytes) -> str:
    """Wrap the filtered value into a ContentBlock of type audio.

    The resulting ChatMessage will have the field `content` populated with a list of ContentBlock objects.

    Supports both file paths and URLs (including data URLs).

    Example:
        ```jinja
        {{ "path/to/audio/file.mp3" | audio }}
        {{ "https://example.com/audio.mp3" | audio }}
        ```
    """
    if isinstance(value, bytes):
        audio_format = _get_audio_format_from_bytes(resolve_binary(value, as_base64=False))
        input_audio = InputAudio.from_bytes(value, audio_format=audio_format)
    elif _is_url(value):
        audio_format = _get_audio_format_from_url(value)
        input_audio = InputAudio.from_url(value, audio_format)
    else:
        input_audio = InputAudio.from_path(Path(value))
    block = ContentBlock.model_validate({"type": "audio", "input_audio": input_audio})
    return f"<content_block>{block.model_dump_json()}</content_block>"



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

from .config import config
from .env import env
from .prompt import AsyncPrompt, Prompt
from .types import ChatMessage

__all__ = ("env", "Prompt", "AsyncPrompt", "config", "ChatMessage")



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import uuid
from typing import Any, Protocol

try:
    from typing import Self
except ImportError:  # pragma: no cover
    from typing_extensions import Self

from jinja2 import meta
from pydantic import BaseModel, ValidationError

from .cache import DefaultCache, RenderCache
from .config import config
from .env import env
from .errors import AsyncError
from .types import ChatMessage, chat_message_from_text
from .utils import generate_canary_word

DEFAULT_VERSION = "0"


class BasePrompt:
    def __init__(
        self,
        text: str,
        *,
        name: str | None = None,
        version: str | None = None,
        metadata: dict[str, Any] | None = None,
        canary_word: str | None = None,
        render_cache: RenderCache | None = None,
    ) -> None:
        """
        Prompt constructor.

        Parameters:
            text: The template text.
            name: The name to identify this prompt.
            version: The version string attached to this prompt.
            metadata: A key-value set of metadata pairs attached to this prompt.
            canary_word: The string to use for the `{{canary_word}}` extension. If `None`, a default string will be
                generated.
            render_cache: The caching backend to store rendered prompts. If `None`, the default in-memory backend will
                be used.
        """
        self._metadata = metadata or {}
        self._name = name or str(uuid.uuid4())
        self._raw: str = text
        self._render_cache = render_cache or DefaultCache()
        self._template = env.from_string(text)
        self._version = version or DEFAULT_VERSION

        self.defaults = {"canary_word": canary_word or generate_canary_word()}

    def _get_context(self, data: dict | None) -> dict:
        if data is None:
            return self.defaults
        return data | self.defaults

    @property
    def metadata(self) -> dict[str, Any]:
        return self._metadata

    @property
    def name(self) -> str:
        return self._name

    @property
    def raw(self) -> str:
        """Returns the raw text of the prompt."""
        return self._raw

    @property
    def version(self) -> str | None:
        return self._version

    @property
    def variables(self) -> set[str]:
        try:
            ast = env.parse(self.raw)
            return meta.find_undeclared_variables(ast)
        except Exception as e:
            from jinja2 import TemplateSyntaxError

            if isinstance(e, TemplateSyntaxError):
                raise
            # Re-raise as TemplateSyntaxError for consistency
            msg = f"Failed to parse template: {e}"
            raise TemplateSyntaxError(msg, 0) from e

    def canary_leaked(self, text: str) -> bool:
        """Returns whether the canary word is present in `text`, signalling the prompt might have leaked."""
        return self.defaults["canary_word"] in text


class Prompt(BasePrompt):
    """
    The `Prompt` class is the only thing you need to know about Banks on the Python side. The class can be
    initialized with a string variable containing the prompt template text, then you can invoke the method `text`
    on your instance to pass the data needed to render the template and get back the final prompt.

    A quick example:
    ```py
    from banks import Prompt


    p = Prompt("Write a 500-word blog post on {{ topic }}.")
    my_topic = "retrogame computing"
    print(p.text({"topic": my_topic}))
    ```
    """

    def text(self, data: dict[str, Any] | None = None) -> str:
        """
        Render the prompt using variables present in `data`

        Parameters:
            data: A dictionary containing the context variables.
        """
        data = self._get_context(data)
        cached = self._render_cache.get(data)
        if cached:
            return cached

        rendered: str = self._template.render(data)
        self._render_cache.set(data, rendered)
        return rendered

    def chat_messages(self, data: dict[str, Any] | None = None) -> list[ChatMessage]:
        """
        Render the prompt using variables present in `data`

        Parameters:
            data: A dictionary containing the context variables.
        """
        data = self._get_context(data)
        rendered = self._render_cache.get(data)
        if not rendered:
            rendered = self._template.render(data)
            self._render_cache.set(data, rendered)

        messages: list[ChatMessage] = []
        for line in rendered.strip().split("\n"):
            try:
                messages.append(ChatMessage.model_validate_json(line))
            except ValidationError:
                # Ignore lines that are not a message
                pass

        if not messages:
            # fallback, if there was no {% chat %} block in the template,
            # try to build a list of messages for the role "user"
            messages.append(chat_message_from_text(role="user", content=rendered))

        return messages


class AsyncPrompt(BasePrompt):
    """
    Banks provides async support through the machinery [provided by Jinja](https://jinja.palletsprojects.com/en/3.0.x/api/#async-support)

    Since the Jinja environment is a global state in banks, the library can work either with or
    without async support, and this must be known before importing anything.

    If the application using banks runs within an `asyncio` loop, you can do two things
    to optimize banks' execution:

    1. Set the environment variable `BANKS_ASYNC_ENABLED=true`.
    2. Use the `AsyncPrompt` class that has an awaitable `run` method.

    For example, let's render a prompt that contains some calls to the `generate` extension. Those calls
    will be heavily I/O bound, so other tasks can take advantage and advance while the prompt is being
    rendered.

    Example:
    ```python
    # Enable async support before importing from banks
    import os

    os.environ["BANKS_ASYNC_ENABLED"] = "true"

    # Show logs to see what's happening at runtime
    import logging

    logging.basicConfig(level=logging.INFO)

    import asyncio
    from banks import AsyncPrompt

    prompt_template = \"\"\"
    Generate a tweet about the topic '{{ topic }}' with a positive sentiment.
    \"\"\"

    async def task(task_id: int, sleep_time: int):
        logging.info(f"Task {task_id} is running.")
        await asyncio.sleep(sleep_time)
        logging.info(f"Task {task_id} done.")


    async def main():
        p = AsyncPrompt(prompt_template)
        # Schedule the prompt rendering along with two executions of 'task', one sleeping for 10 seconds
        # and one sleeping for 1 second
        results = await asyncio.gather(p.text({"topic": "AI frameworks"}), task(1, 10), task(2, 1))
        print("All tasks done, rendered prompt:")
        print(results[0])


    asyncio.run(main())
    ```
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        if not config.ASYNC_ENABLED:
            msg = "Async is not enabled. Please set the environment variable 'BANKS_ASYNC_ENABLED=on' and try again."
            raise AsyncError(msg)

    async def text(self, data: dict[str, Any] | None = None) -> str:
        """
        Render the prompt using variables present in `data`

        Parameters:
            data: A dictionary containing the context variables.
        """
        data = self._get_context(data)
        cached = self._render_cache.get(data)
        if cached:
            return cached

        rendered: str = await self._template.render_async(data)
        self._render_cache.set(data, rendered)
        return rendered


class PromptRegistry(Protocol):  # pragma: no cover
    """Interface to be implemented by concrete prompt registries."""

    def get(self, *, name: str, version: str | None = None) -> Prompt: ...

    def set(self, *, prompt: Prompt, overwrite: bool = False) -> None: ...


class PromptModel(BaseModel):
    """Serializable representation of a Prompt."""

    text: str
    name: str | None = None
    version: str | None = None
    metadata: dict[str, Any] | None = None

    @classmethod
    def from_prompt(cls: type[Self], prompt: Prompt) -> Self:
        return cls(text=prompt.raw, name=prompt.name, version=prompt.version, metadata=prompt.metadata)



# banks/tests/test_redis_registry.py
import pytest
import redis

from banks.errors import InvalidPromptError, PromptNotFoundError
from banks.prompt import Prompt
from banks.registries.redis import RedisPromptRegistry


@pytest.fixture
def redis_client():
    client = redis.Redis(host="localhost", port=6379, db=0)
    # Clear test database before each test
    client.flushdb()
    return client


@pytest.fixture
def registry(redis_client):  # type: ignore[ARG001]
    return RedisPromptRegistry(redis_url="redis://localhost:6379")


@pytest.mark.redis
def test_get_not_found(registry):
    with pytest.raises(PromptNotFoundError):
        registry.get(name="nonexistent")


@pytest.mark.redis
def test_set_and_get_prompt(registry):
    prompt = Prompt("Hello {{name}}!", name="greeting")
    registry.set(prompt=prompt)

    retrieved = registry.get(name="greeting")
    assert retrieved.raw == "Hello {{name}}!"
    assert retrieved.name == "greeting"
    assert retrieved.version == "0"  # default version
    assert retrieved.metadata == {}


@pytest.mark.redis
def test_set_existing_no_overwrite(registry):
    prompt = Prompt("Hello {{name}}!", name="greeting")
    registry.set(prompt=prompt)

    new_prompt = Prompt("Hi {{name}}!", name="greeting")
    with pytest.raises(
        InvalidPromptError, match=r"Prompt with name 'greeting' already exists. Use overwrite=True to overwrite"
    ):
        registry.set(prompt=new_prompt)


@pytest.mark.redis
def test_set_existing_overwrite(registry):
    prompt = Prompt("Hello {{name}}!", name="greeting")
    registry.set(prompt=prompt)

    new_prompt = Prompt("Hi {{name}}!", name="greeting")
    registry.set(prompt=new_prompt, overwrite=True)

    retrieved = registry.get(name="greeting")
    assert retrieved.raw == "Hi {{name}}!"


@pytest.mark.redis
def test_set_multiple_versions(registry):
    prompt_v1 = Prompt("Version 1", name="multi", version="1")
    prompt_v2 = Prompt("Version 2", name="multi", version="2")

    registry.set(prompt=prompt_v1)
    registry.set(prompt=prompt_v2)

    retrieved_v1 = registry.get(name="multi", version="1")
    assert retrieved_v1.raw == "Version 1"

    retrieved_v2 = registry.get(name="multi", version="2")
    assert retrieved_v2.raw == "Version 2"


@pytest.mark.redis
def test_get_with_version(registry):
    prompt = Prompt("Test {{var}}", name="test", version="1.0")
    registry.set(prompt=prompt)

    retrieved = registry.get(name="test", version="1.0")
    assert retrieved.version == "1.0"
    assert retrieved.raw == "Test {{var}}"


@pytest.mark.redis
def test_set_with_metadata(registry):
    prompt = Prompt("Test prompt", name="test", metadata={"author": "John", "category": "test"})
    registry.set(prompt=prompt)

    retrieved = registry.get(name="test")
    assert retrieved.metadata == {"author": "John", "category": "test"}


@pytest.mark.redis
def test_update_metadata(registry):
    # Initial prompt with metadata
    prompt = Prompt("Test prompt", name="test", metadata={"score": 90})
    registry.set(prompt=prompt)

    # Update metadata
    updated_prompt = registry.get(name="test")
    updated_prompt.metadata["score"] = 95
    registry.set(prompt=updated_prompt, overwrite=True)

    # Verify update
    retrieved = registry.get(name="test")
    assert retrieved.metadata["score"] == 95


@pytest.mark.redis
def test_invalid_redis_connection():
    with pytest.raises(redis.ConnectionError):
        registry = RedisPromptRegistry(redis_url="redis://nonexistent:6379")
        registry.get(name="test")


@pytest.mark.redis
def test_custom_prefix(redis_client):
    registry = RedisPromptRegistry(redis_url="redis://localhost:6379", prefix="custom:prefix:")

    prompt = Prompt("Test", name="test")
    registry.set(prompt=prompt)

    # Verify the key in Redis has the custom prefix
    key = "custom:prefix:test:0"
    assert redis_client.exists(key)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import re
from pathlib import Path
from urllib.parse import urlparse

from banks.types import ContentBlock, ImageUrl

BASE64_PATH_REGEX = re.compile(r"image\/.*;base64,.*")


def _is_url(string: str) -> bool:
    result = urlparse(string)
    if not result.scheme:
        return False

    if not result.netloc:
        # The only valid format when netloc is empty is base64 data urls
        return all([result.scheme == "data", BASE64_PATH_REGEX.match(result.path)])

    return True


def image(value: str | bytes) -> str:
    """Wrap the filtered value into a ContentBlock of type image.

    The resulting ChatMessage will have the field `content` populated with a list of ContentBlock objects.

    Example:
        ```jinja
        Describe what you see

        {{ "path/to/image/file" | image }}
        ```

    Important:
        this filter marks the content to cache by surrounding it with `<content_block>` and
        `</content_block>`, so it's only useful when used within a `{% chat %}` block.
    """
    if isinstance(value, bytes):
        image_url = ImageUrl.from_bytes(bytes_str=value)
    elif _is_url(value):
        image_url = ImageUrl(url=value)
    else:
        image_url = ImageUrl.from_path(Path(value))

    block = ContentBlock.model_validate({"type": "image_url", "image_url": image_url})
    return f"<content_block>{block.model_dump_json()}</content_block>"



import xml.etree.ElementTree as ET
from typing import Any
from xml.dom.minidom import parseString

import pytest
from pydantic import BaseModel

from banks.filters.xml import xml


@pytest.fixture
def xml_string_from_basemodel() -> str:
    model = ET.Element("person")
    age = ET.SubElement(model, "age")
    age.text = "30"
    name = ET.SubElement(model, "name")
    name.text = "John Doe"
    xml_str = ET.tostring(model, encoding="unicode")
    return parseString(xml_str).toprettyxml().replace('<?xml version="1.0" ?>\n', "")  # noqa: S318


@pytest.fixture
def xml_string_from_other() -> str:
    model = ET.Element("input")
    age = ET.SubElement(model, "age")
    age.text = "30"
    name = ET.SubElement(model, "name")
    name.text = "John Doe"
    xml_str = ET.tostring(model, encoding="unicode")
    return parseString(xml_str).toprettyxml().replace('<?xml version="1.0" ?>\n', "")  # noqa: S318


@pytest.fixture
def xml_string_from_edge_case() -> str:
    model = ET.Element("input")
    hello = ET.SubElement(model, "_1")
    hello.text = "hello"
    world = ET.SubElement(model, "_2.4")
    world.text = "world"
    xml_str = ET.tostring(model, encoding="unicode")
    return parseString(xml_str).toprettyxml().replace('<?xml version="1.0" ?>\n', "")  # noqa: S318


@pytest.fixture
def starting_value() -> tuple[BaseModel, dict[str, Any], str]:
    class Person(BaseModel):
        age: int
        name: str

    p = Person(age=30, name="John Doe")
    return p, p.model_dump(), p.model_dump_json()


@pytest.fixture
def dict_edge_case() -> tuple[dict, str]:
    return {1: "hello", 2.4: "world"}


def test_xml_filter(
    xml_string_from_basemodel: str,
    xml_string_from_other: str,
    starting_value: tuple[BaseModel, dict[str, Any], str],
    dict_edge_case: dict,
    xml_string_from_edge_case: str,
) -> None:
    model, dictionary, string = starting_value
    assert xml(model) == xml_string_from_basemodel
    assert xml(dictionary) == xml_string_from_other
    assert xml(string) == xml_string_from_other
    assert xml(dict_edge_case) == xml_string_from_edge_case



from __future__ import annotations

import json
from typing import cast

from banks import Prompt
from banks.errors import InvalidPromptError, PromptNotFoundError
from banks.prompt import DEFAULT_VERSION, PromptModel

REDIS_INSTALL_MSG = "redis is not installed. Please install it with `pip install redis`."


class RedisPromptRegistry:
    """A prompt registry that stores prompts in Redis."""

    def __init__(
        self,
        redis_url: str = "redis://localhost:6379",
        prefix: str = "banks:prompt:",
    ) -> None:
        """
        Initialize the Redis prompt registry.

        Parameters:
            redis_url: Redis connection URL
            prefix: Key prefix for storing prompts in Redis
        """
        try:
            import redis
        except ImportError as e:
            raise ImportError(REDIS_INSTALL_MSG) from e

        self._redis = redis.from_url(redis_url, decode_responses=True)
        self._prefix = prefix

    def _make_key(self, name: str, version: str) -> str:
        """Create Redis key for a prompt."""
        return f"{self._prefix}{name}:{version}"

    def get(self, *, name: str, version: str | None = None) -> Prompt:
        """
        Get a prompt by name and version.

        Parameters:
            name: Name of the prompt
            version: Version of the prompt (optional)

        Returns:
            Prompt instance

        Raises:
            PromptNotFoundError: If prompt doesn't exist
        """
        version = version or DEFAULT_VERSION
        key = self._make_key(name, version)

        data = self._redis.get(key)
        if not data:
            msg = f"Cannot find prompt with name '{name}' and version '{version}'"
            raise PromptNotFoundError(msg)

        prompt_data = json.loads(cast(str, data))
        return Prompt(**prompt_data)

    def set(self, *, prompt: Prompt, overwrite: bool = False) -> None:
        """
        Store a prompt in Redis.

        Parameters:
            prompt: Prompt instance to store
            overwrite: Whether to overwrite existing prompt

        Raises:
            InvalidPromptError: If prompt exists and overwrite=False
        """
        version = prompt.version or DEFAULT_VERSION
        key = self._make_key(prompt.name, version)

        # Check if prompt already exists
        if self._redis.exists(key) and not overwrite:
            msg = f"Prompt with name '{prompt.name}' already exists. Use overwrite=True to overwrite"
            raise InvalidPromptError(msg)

        # Convert prompt to serializable format
        prompt_model = PromptModel.from_prompt(prompt)
        prompt_data = prompt_model.model_dump()

        # Store in Redis
        self._redis.set(key, json.dumps(prompt_data))



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import base64
import re
from base64 import b64decode, b64encode
from binascii import Error as BinasciiError
from enum import Enum
from inspect import Parameter, getdoc, signature
from pathlib import Path
from typing import Callable, Literal, Union, cast

import filetype  # type: ignore[import-untyped]
from pydantic import BaseModel
from typing_extensions import Self

from .utils import parse_params_from_docstring, python_type_to_jsonschema

# pylint: disable=invalid-name
CONTENT_BLOCK_REGEX = re.compile(r"(<content_block>\{.*?\}<\/content_block>)|([^<](?:(?!<content_block>)[\s\S])*)")


def resolve_binary(bytes_str: bytes, *, as_base64: bool = True) -> bytes:
    """
    Resolve binary data between base64 and raw bytes.

    Args:
        bytes_str: Bytes data
        as_base64: Whether to return base64 encoded bytes or raw bytes

    Returns:
        b64 encoded bytes if input is not base64 encoded, else returns input as is.
    """
    # check if bytes_str is base64 encoded
    try:
        # Check if raw_bytes is already base64 encoded.
        # b64decode() can succeed on random binary data, so we
        # pass verify=True to make sure it's not a false positive
        raw_bytes = base64.b64decode(bytes_str, validate=True)
        b64_bytes = bytes_str
    except BinasciiError:
        # b64decode failed, leave as is
        raw_bytes = bytes_str
        b64_bytes = b64encode(bytes_str)
    return b64_bytes if as_base64 else raw_bytes


class ContentBlockType(str, Enum):
    text = "text"
    image_url = "image_url"
    audio = "audio"
    video = "video"
    document = "document"


class CacheControl(BaseModel):
    type: str = "ephemeral"


class ImageUrl(BaseModel):
    url: str

    @staticmethod
    def _mimetype_from_bytes(raw_bytes: bytes) -> str:
        kind = filetype.guess(raw_bytes)
        if kind is not None:
            return kind.mime
        # Default to jpeg if format cannot be determined
        return "image/jpeg"

    @classmethod
    def from_base64(cls, media_type: str, base64_str: str) -> Self:
        return cls(url=f"data:{media_type};base64,{base64_str}")

    @classmethod
    def from_path(cls, file_path: Path) -> Self:
        with open(file_path, "rb") as image_file:
            raw_bytes = image_file.read()
            mimetype = cls._mimetype_from_bytes(raw_bytes)
            return cls.from_base64(mimetype, base64.b64encode(raw_bytes).decode("utf-8"))

    @classmethod
    def from_bytes(cls, bytes_str: bytes) -> Self:
        """Create ImageUrl from bytes
        Args:
            bytes_str: Bytes data
        Returns:
            ImageUrl instance with base64 encoded bytes as URL
        """
        b64_bytes = resolve_binary(bytes_str)
        mimetype = cls._mimetype_from_bytes(b64decode(b64_bytes))
        return cls.from_base64(mimetype, b64_bytes.decode("utf-8"))


AudioFormat = Literal["mp3", "wav", "m4a", "webm", "ogg", "flac"]
VideoFormat = Literal["mp4", "mpg", "mov", "avi", "flv", "webm", "wmv", "3gp", "3gpp"]
# Because Claude only supports pdf and text, and Gemini only supports a small subset of text formats,
# we can default to 'txt' for any text-based format that is not pdf. This allows the data to be sent to the llm
# in an acceptable format, but the LLM should still be able to understand the content: e.g., html, markdown,
# xml, etc.
# If detecting text types should become desirable, I recommend using something like Google magicka
DocumentFormat = Literal["pdf", "txt"]


class InputAudio(BaseModel):
    data: str
    format: AudioFormat

    @classmethod
    def from_path(cls, file_path: Path) -> Self:
        with open(file_path, "rb") as audio_file:
            encoded_str = base64.b64encode(audio_file.read()).decode("utf-8")
            file_format = cast(AudioFormat, file_path.suffix[1:])
            return cls(data=encoded_str, format=file_format)

    @classmethod
    def from_url(cls, url: str, audio_format: AudioFormat) -> Self:
        """Create InputAudio from a URL.

        Args:
            url: The URL to the audio file
            audio_format: The audio format

        Returns:
            InputAudio instance with the URL as data
        """
        return cls(data=url, format=audio_format)

    @classmethod
    def from_bytes(cls, bytes_str: bytes, audio_format: AudioFormat) -> Self:
        """Create InputAudio from bytes

        Args:
            bytes_str: Bytes data
            audio_format: The audio format

        Returns:
            InputAudio instance with base64 encoded bytes as data
        """
        b64_bytes = resolve_binary(bytes_str)
        encoded_str = b64_bytes.decode("utf-8")
        return cls(data=encoded_str, format=audio_format)


class InputVideo(BaseModel):
    data: str
    format: VideoFormat

    @classmethod
    def from_path(cls, file_path: Path) -> Self:
        with open(file_path, "rb") as video_file:
            encoded_str = base64.b64encode(video_file.read()).decode("utf-8")
            file_format = cast(VideoFormat, file_path.suffix[1:])
            return cls(data=encoded_str, format=file_format)

    @classmethod
    def from_url(cls, url: str, video_format: VideoFormat) -> Self:
        """Create InputVideo from a URL.

        Args:
            url: The URL to the audio file
            video_format: The audio format

        Returns:
            InputVideo instance with the URL as data
        """
        return cls(data=url, format=video_format)

    @classmethod
    def from_bytes(cls, bytes_str: bytes, video_format: VideoFormat) -> Self:
        """Create InputVideo from bytes

        Args:
            bytes_str: Bytes data
            video_format: The video format

        Returns:
            InputVideo instance with base64 encoded bytes as data
        """
        b64_bytes = resolve_binary(bytes_str)
        encoded_str = b64_bytes.decode("utf-8")
        return cls(data=encoded_str, format=video_format)


class InputDocument(BaseModel):
    data: str
    format: DocumentFormat

    @classmethod
    def from_path(cls, file_path: Path) -> Self:
        with open(file_path, "rb") as document_file:
            encoded_str = base64.b64encode(document_file.read()).decode("utf-8")
            file_format = cast(DocumentFormat, file_path.suffix[1:])
            return cls(data=encoded_str, format=file_format)

    @classmethod
    def from_url(cls, url: str, document_format: DocumentFormat) -> Self:
        """Create InputDocument from a URL.

        Args:
            url: The URL to the document file
            document_format: The document format

        Returns:
            InputDocument instance with the URL as data
        """
        return cls(data=url, format=document_format)

    @classmethod
    def from_bytes(cls, bytes_str: bytes, document_format: DocumentFormat) -> Self:
        """Create InputDocument from bytes

        Args:
            bytes_str: Bytes data
            document_format: The document format

        Returns:
            InputDocument instance with base64 encoded bytes as data
        """
        b64_bytes = resolve_binary(bytes_str)
        encoded_str = b64_bytes.decode("utf-8")
        return cls(data=encoded_str, format=document_format)


class ContentBlock(BaseModel):
    type: ContentBlockType
    cache_control: CacheControl | None = None
    text: str | None = None
    image_url: ImageUrl | None = None
    input_audio: InputAudio | None = None
    input_video: InputVideo | None = None
    input_document: InputDocument | None = None


ChatMessageContent = Union[list[ContentBlock], str]


class ChatMessage(BaseModel):
    role: str
    content: ChatMessageContent
    tool_call_id: str | None = None
    name: str | None = None


class FunctionParameter(BaseModel):
    type: str
    description: str


class FunctionParameters(BaseModel):
    type: str = "object"
    properties: dict[str, FunctionParameter]
    required: list[str]


class Function(BaseModel):
    name: str
    description: str
    parameters: FunctionParameters


class Tool(BaseModel):
    """A model representing a Tool to be used in function calling.

    This model should dump the following:
    ```
    {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                },
                "required": ["location"],
            },
        },
        "import_path": "module.get_current_weather",
    }
    ```
    """

    type: str = "function"
    function: Function
    import_path: str

    @classmethod
    def from_callable(cls, func: Callable) -> Self:
        sig = signature(func)

        # Try getting params descriptions from docstrings
        param_docs = parse_params_from_docstring(func.__doc__ or "")

        # If the docstring is missing, use the qualname space-separated hopefully
        # the LLM will get some more context than just using the function name.
        description = getdoc(func) or " ".join(func.__qualname__.split("."))
        properties = {}
        required = []
        for name, param in sig.parameters.items():
            p = FunctionParameter(
                description=param_docs.get(name, {}).get("description", ""),
                type=python_type_to_jsonschema(param.annotation),
            )
            properties[name] = p
            if param.default == Parameter.empty:
                required.append(name)

        return cls(
            function=Function(
                name=func.__name__,
                description=description,
                parameters=FunctionParameters(properties=properties, required=required),
            ),
            import_path=f"{func.__module__}.{func.__qualname__}",
        )


def chat_message_from_text(role: str, content: str) -> ChatMessage:
    """
    Helper callback.
    """
    content_blocks: list[ContentBlock] = []

    # Find all content block matches
    matches = CONTENT_BLOCK_REGEX.finditer(content)
    for match in matches:
        if match.group(1):
            # content block match
            content_block_json_str = (
                match.group(1).strip().removeprefix("<content_block>").removesuffix("</content_block>")
            )
            content_blocks.append(ContentBlock.model_validate_json(content_block_json_str))
        elif match.group(2):
            # plain-text match
            text = match.group(2).strip()
            if text:
                content_blocks.append(ContentBlock(type=ContentBlockType.text, text=text))

    # If no content blocks were found, treat entire content as text
    if not content_blocks:
        content_blocks.append(ContentBlock(type=ContentBlockType.text, text=content))

    final_content = content_blocks

    return ChatMessage(role=role, content=final_content)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from banks.types import ContentBlock


def cache_control(value: str, cache_type: str = "ephemeral") -> str:
    """Wrap the filtered value into a ContentBlock with the proper cache_control field set.

    The resulting ChatMessage will have the field `content` populated with a list of ContentBlock objects.

    Example:
        ```jinja
        {{ "This is a long, long text" | cache_control("ephemeral") }}

        This is short and won't be cached.
        ```

    Important:
        this filter marks the content to cache by surrounding it with `<content_block>` and
        `</content_block>`, so it's only useful when used within a `{% chat %}` block.
    """
    block = ContentBlock.model_validate({"type": "text", "text": value, "cache_control": {"type": cache_type}})
    return f"<content_block>{block.model_dump_json()}</content_block>"



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from .audio import audio
from .cache_control import cache_control
from .document import document
from .image import image
from .lemmatize import lemmatize
from .tool import tool
from .video import video
from .xml import xml

__all__ = ("cache_control", "image", "lemmatize", "tool", "audio", "video", "document", "xml")



import json
from base64 import b64encode
from pathlib import Path

import pytest

from banks import Prompt
from banks.filters.document import _get_document_format_from_bytes, _get_document_format_from_url, _is_url, document


@pytest.fixture
def tiny_pdf():
    here = Path(__file__).parent
    return here / "data" / "1x1.pdf"


def test_document_with_file_path(tiny_pdf):
    """Test document filter with a file path input"""
    result = document(str(tiny_pdf))

    # Verify the content block wrapper
    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    # Parse the JSON content
    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "document"
    assert content_block["input_document"]["format"].startswith("pdf")


def test_document_with_nonexistent_file():
    """Test document filter with a nonexistent file path"""
    with pytest.raises(FileNotFoundError):
        document("nonexistent/document.pdf")


def test_document_with_url():
    """Test document filter with a URL input (no filesystem access)."""
    url = "https://example.com/document.css"
    result = document(url)

    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    # All text types are treated as txt
    assert content_block["type"] == "document"
    assert content_block["input_document"]["data"] == url
    assert content_block["input_document"]["format"] == "txt"


def test_is_url_variants():
    assert _is_url("relative/path.pdf") is False
    assert _is_url("https://example.com/document.pdf") is True
    assert _is_url("data:application/pdf;base64,AAAA") is True
    assert _is_url("data:text/plain;base64,AAAA") is True
    assert _is_url("data:audio/mp3;base64,AAAA") is False


def test_get_document_format_from_url():
    with pytest.raises(ValueError):
        _get_document_format_from_url("https://example.com/document.WAV")
    assert _get_document_format_from_url("https://example.com/document") == "pdf"


def test_get_document_from_bytes(tiny_pdf):
    # PDF file header bytes
    with open(tiny_pdf, "rb") as f:
        pdf_bytes = f.read()
    result = document(pdf_bytes)

    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    assert content_block["type"] == "document"
    assert content_block["input_document"]["format"] == "pdf"
    assert content_block["input_document"]["data"] == b64encode(pdf_bytes).decode()


def test_get_document_from_b64_bytes():
    html_str = "<html><body><h1>Test Document</h1></body></html>"
    html_bytes = html_str.encode("utf-8")
    html_b64 = b64encode(html_bytes)
    result = document(html_b64)

    assert result.startswith("<content_block>")
    assert result.endswith("</content_block>")

    json_content = result[15:-16]  # Remove wrapper tags
    content_block = json.loads(json_content)

    # All non pdf documents are treated as txt
    assert content_block["type"] == "document"
    assert content_block["input_document"]["format"] == "txt"
    assert content_block["input_document"]["data"] == html_b64.decode()


def test_get_document_format_from_bytes(tiny_pdf):
    with open(tiny_pdf, "rb") as f:
        pdf_bytes = f.read()
    html_bytes = b"<html><body><h1>Test</h1></body></html>"
    xhtml_bytes = (
        b"<!DOCTYPE html>"
        b'<html xmlns="http://www.w3.org/1999/xhtml"><head><title>Test</title></head><body></body>'
        b"</html>"
    )
    css_bytes = (
        b"body {\n  margin: 0;\n  font-family: Arial, sans-serif;\n  background: #fff;\n  color: #222;\n}\n\n"
        b".btn {\n  display: inline-block;\n  padding: 8px 12px;\n  background: #007bff;\n  color: #fff;\n  border-radius: 4px;\n  text-decoration: none;\n}\n\n"
        b".btn:hover {\n  background: #0056b3;\n}"
    )
    txt_bytes = b"This is a plain text document.\nIt has multiple lines.\nEnd of document."
    md_bytes = (
        b"# This is a Markdown document\n\n"
        b"This document contains **bold** text and *italic* text."
        b"\n\n"
        b"- Item 1\n"
        b"- Item 2\n"
        b"- Item 3\n"
    )
    rst_bytes = (
        b"This is a reStructuredText document.\n"
        b"===============================\n\n"
        b"This document contains **bold** text and *italic* text.\n\n"
        b"- Item 1\n"
        b"- Item 2\n"
        b"- Item 3\n"
    )
    xml_bytes = b'<?xml version="1.0" encoding="UTF-8"?><note><to>User</to><from>Tester</from><heading>Reminder</heading><body>This is a test XML document.</body></note>'
    csv = b"Name,Age,Location\nAlice,30,New York\nBob,25,Los Angeles\nCharlie,35,Chicago\n"
    rtf = b"{\\rtf1\\ansi\\deff0 {\\fonttbl {\\f0 Arial;}}\\f0\\fs24 This is a simple RTF document.}"
    js = b"function greet(name) { console.log('Hello, ' + name + '!'); } greet('World');"
    mjs = b"export function add(a, b) { return a + b; } console.log(add(2, 3));"
    cjs = b"const fs = require('fs'); fs.readFile('example.txt', 'utf8', (err, data) => { if (err) throw err; console.log(data); });"
    json_bytes = b'{ "name": "John", "age": 30, "city": "New York" }'

    assert _get_document_format_from_bytes(pdf_bytes) == "pdf"
    assert _get_document_format_from_bytes(html_bytes) == "txt"
    assert _get_document_format_from_bytes(xhtml_bytes) == "txt"
    assert _get_document_format_from_bytes(css_bytes) == "txt"
    assert _get_document_format_from_bytes(txt_bytes) == "txt"
    assert _get_document_format_from_bytes(md_bytes) == "txt"
    assert _get_document_format_from_bytes(rst_bytes) == "txt"
    assert _get_document_format_from_bytes(xml_bytes) == "txt"
    assert _get_document_format_from_bytes(csv) == "txt"
    assert _get_document_format_from_bytes(rtf) == "txt"
    assert _get_document_format_from_bytes(js) == "txt"
    assert _get_document_format_from_bytes(mjs) == "txt"
    assert _get_document_format_from_bytes(cjs) == "txt"
    assert _get_document_format_from_bytes(json_bytes) == "txt"


def test_document_no_chat_block(tiny_pdf):
    prompt = Prompt("{{ test }} and {{ another | document }}")
    messages = prompt.chat_messages({"test": "hello world", "another": str(tiny_pdf)})
    assert len(messages) == 1
    message = messages[0]
    assert len(message.content) == 2
    assert message.content[0].text == "hello world and"  # type: ignore
    assert message.content[1].type == "document"  # type:ignore



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import mimetypes
import re
from pathlib import Path
from typing import cast
from urllib.parse import urlparse

import filetype  # type: ignore[import-untyped]

from banks.types import ContentBlock, DocumentFormat, InputDocument, resolve_binary

BASE64_DOCUMENT_REGEX = re.compile(r"(text|application)\/.*;base64,.*")


def _is_url(string: str) -> bool:
    """Check if a string is a URL."""
    result = urlparse(string)
    if not result.scheme:
        return False

    if not result.netloc:
        # The only valid format when netloc is empty is base64 data urls
        return all([result.scheme == "data", BASE64_DOCUMENT_REGEX.match(result.path)])

    return True


def _get_document_format_from_url(url: str) -> DocumentFormat:
    """Extract document format from URL.

    Tries to determine format from URL path or defaults to pdf.
    """
    parsed = urlparse(url)
    path = parsed.path.lower()
    # Gemini supported file types https://ai.google.dev/gemini-api/docs/file-input-methods
    # text/html
    # text/css
    # text/plain
    # text/xml
    # text/csv
    # text/rtf
    # text/javascript
    # application/json
    # application/pdf

    # Claude supported file types
    # application/pdf
    # text/plain

    # OpenAI supported file types
    # application/pdf

    for fmt in (
        "pdf",
        "html",
        "htm",
        "xhtml",
        "css",
        "txt",
        "md",
        "markdown",
        "rst",
        "xml",
        "csv",
        "rtf",
        "js",
        "mjs",
        "cjs",
        "javascript",
        "json",
    ):
        # Because Claude only supports pdf and text, and Gemini only supports a small subset of text formats,
        # we can default to 'txt' for any text-based format that is not pdf. This allows the data to be sent to the llm
        # in an acceptable format, but the LLM should still be able to understand the content: e.g., html, markdown,
        # xml, etc.
        if path.endswith(f".{fmt}"):
            if fmt == "pdf":
                return cast(DocumentFormat, "pdf")
            return "txt"
    mime = mimetypes.guess_type(path)[0]
    if mime is not None and mime.startswith("text/"):
        return "txt"
    # With urls, the likelihood seems sufficiently high that it's probably a pdf if not otherwise indicated
    if mime is None:
        return "pdf"
    # Document type indicated to be other than pdf or text type
    raise ValueError("Unsupported document format: " + path)


def _get_document_format_from_bytes(data: bytes) -> DocumentFormat:
    """Extract document format from bytes data using filetype library."""
    # First check for pdf (only non text based format) and RTF formats (can be detected by file header)
    kind = filetype.guess(data)
    if kind is not None:
        fmt = kind.extension
        if fmt == "pdf":
            return cast(DocumentFormat, fmt)

    # filetype is good at detecting binary formats, but not text-based ones.
    # So, this is a good indicator that it's text-based.
    # Because Claude only supports pdf and text, and Gemini only supports a small subset of text formats,
    # we can default to 'txt' for any text-based format that is not pdf. This allows the data to be sent to the llm in
    # an acceptable format, but the LLM should still be able to understand the content: e.g., html, markdown, xml, etc.
    # If detecting text types should become desirable, I recommend using something like Google magicka
    if kind is None or kind.extension == "rtf":
        return "txt"
    # There are many common document types (like word, excel, powerpoint, etc.) that are not supported.
    raise ValueError("Unsupported document format: " + kind.extension)


def document(value: str | bytes) -> str:
    """Wrap the filtered value into a ContentBlock of type document.

    The resulting ChatMessage will have the field `content` populated with a list of ContentBlock objects.

    Supports both file paths and URLs (including data URLs).

    Example:
        ```jinja
        {{ "path/to/document/file.pdf" | document }}
        {{ "https://example.com/document.pdf" | document }}
        ```
    """
    if isinstance(value, bytes):
        document_format = _get_document_format_from_bytes(resolve_binary(value, as_base64=False))
        input_document = InputDocument.from_bytes(value, document_format=document_format)
    elif _is_url(value):
        document_format = _get_document_format_from_url(value)
        input_document = InputDocument.from_url(value, document_format)
    else:
        input_document = InputDocument.from_path(Path(value))
    block = ContentBlock.model_validate({"type": "document", "input_document": input_document})
    return f"<content_block>{block.model_dump_json()}</content_block>"
